#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "umfpack.h"
#include <petscksp.h>
#include <stdbool.h>

double getTime()
{
  struct timeval time;
  gettimeofday(&time, NULL);
  double t = (double)time.tv_sec + 1e-6*(double)time.tv_usec;

  return t;
}

//-----------------------------------------------------------------------------
// Compressed Sparse Column format matrix
struct _CSC
{
  int m;  // matrix is m x m
  int len_nzval;
  int* colptr;
  int* rowval;
  double* nzval;
};

typedef struct _CSC CSC;

CSC makeCSC(int m, int len_nzval)
{
  int* colptr = (int*)malloc(sizeof(int)*(m + 1));
  int* rowval = (int*)malloc(sizeof(int)*len_nzval);
  double* nzval = (double*)malloc(sizeof(double)*len_nzval);

  CSC A = {m, len_nzval, colptr, rowval, nzval};
  return A;
}

void freeCSC(CSC A)
{
  free(A.colptr);
  free(A.rowval);
  free(A.nzval);
}

//-----------------------------------------------------------------------------
// Read matrix from file

typedef long int mint;  // integer size of data in the files
const size_t len_colptr = 455672 + 1;
const size_t len_nzval = 15079424;

CSC readMatrix()
{
  // read integer arrays with integer mint
  mint* colptr = (mint*)malloc(sizeof(mint)*len_colptr);
  mint* rowval = (mint*)malloc(sizeof(mint)*len_nzval);
  CSC A = makeCSC(len_colptr - 1, len_nzval);

  FILE* f = fopen("colptr.dat", "r");
  fread(colptr, sizeof(mint), len_colptr, f);
  fclose(f);

  f = fopen("rowval.dat", "r");
  fread(rowval, sizeof(mint), len_nzval, f);
  fclose(f);

  f = fopen("nzval.dat", "r");
  fread(A.nzval, sizeof(double), len_nzval, f);
  fclose(f);

  // copy into arrays of int, also make indices zero-based
  for (int i=0; i < len_colptr; i++)
    A.colptr[i] = colptr[i] - 1;

  for (int i=0; i < len_nzval; i++)
    A.rowval[i] = rowval[i] - 1;

  free(colptr); free(rowval);

  return A;
}


//-----------------------------------------------------------------------------
// Solve with Suitesparse (UMFPACK)

void solve_suitesparse()
{
  double* null = (double*)NULL;
  void *Symbolic, *Numeric;
  printf("reading matrix...");
  CSC A = readMatrix();
  printf("finished\n");

  double* x = (double*)malloc(sizeof(double)*A.m);
  double* b = (double*)malloc(sizeof(double)*A.m);
  for (int i=0; i < A.m; i++)
    b[i] = 0.01;

  double tstart, tend, t_totalstart, t_totalend;

  t_totalstart = getTime();

  printf("symbolic factorization...");
  tstart = getTime();
  umfpack_di_symbolic(A.m, A.m, A.colptr, A.rowval, A.nzval, &Symbolic, null,
                      null);
  tend = getTime();
  printf("finished (%e seconds)\n", tend - tstart);


  printf("numeric factorization...");
  tstart = getTime();
  umfpack_di_numeric(A.colptr, A.rowval, A.nzval, Symbolic, &Numeric, null,
                     null);
  tend = getTime();
  printf("finished (%e seconds)\n", tend - tstart);
  umfpack_di_free_symbolic(&Symbolic);

  printf("backsolve...");
  tstart = getTime();
  umfpack_di_solve(UMFPACK_A, A.colptr, A.rowval, A.nzval, x, b, Numeric, null,
                   null);
  tend = getTime();
  printf("finished (%e seconds)\n", tend - tstart);
  umfpack_di_free_numeric(&Numeric);

  t_totalend = getTime();
  printf("total elapsed time: %e seconds\n", t_totalend - t_totalstart);

  freeCSC(A); free(x); free(b);
}


//-----------------------------------------------------------------------------
// Solve with Petsc

Mat createPetscMat()
{
  printf("reading matrix...");
  CSC A = readMatrix();
  printf("finished\n");

  double t_start, t_end;

  // make a Petsc Matrix out of the CSC matrix
  Mat Ap;
  MatCreate(PETSC_COMM_WORLD, &Ap);
  MatSetSizes(Ap, PETSC_DECIDE, PETSC_DECIDE, A.m, A.m);
  MatSetFromOptions(Ap);

  // set preallocation
  printf("preallocating matrix...");
  t_start = getTime();
  PetscInt* dnnz = (PetscInt*)malloc(sizeof(int)*A.m);
  PetscInt* onnz = (PetscInt*)malloc(sizeof(int)*A.m);
  for (int i=0; i < A.m; i++)
  {
    // the matrix is structurally symmetric, so even though A is CSC, we can
    // still use the sparsity counts for a CSR matrix
    dnnz[i] = A.colptr[i+1] - A.colptr[i];
    onnz[i] = 0;
  }
  MatXAIJSetPreallocation(Ap, 1, dnnz, onnz, NULL, NULL);
  t_end = getTime();
  printf("finished (%e seconds)\n", t_end - t_start);

  printf("copying values...");
  t_start = getTime();
  // copy values from A to Ap
  // doing this one entry at a time is inefficient, but we are not timing the
  // matrix assembly
  for (int j=0; j < A.m; j++)
  {
    bool found_diag = false;
    for (int i=A.colptr[j]; i < A.colptr[j+1]; i++)
    {
      MatSetValue(Ap, A.rowval[i], j, A.nzval[i], ADD_VALUES);
      if (j == A.rowval[i])
        found_diag = true;
    }

    if (!found_diag)
      printf("missing diagonal for row %d\n", j);
  }

  MatAssemblyBegin(Ap, MAT_FINAL_ASSEMBLY);
  MatAssemblyEnd(Ap, MAT_FINAL_ASSEMBLY);
  t_end = getTime();
  printf("finished (%e seconds)\n", t_end - t_start);

  freeCSC(A); free(dnnz); free(onnz);

  return Ap;
}


void solve_petsc()
{

  PetscInt m;
  Mat A = createPetscMat();
  MatGetSize(A, &m, NULL);

  // set up vectors
  Vec x, b;
  VecCreate(PETSC_COMM_WORLD, &x);
  VecCreate(PETSC_COMM_WORLD, &b);
  VecSetSizes(x, PETSC_DECIDE, m);
  VecSetSizes(b, PETSC_DECIDE, m);
  VecSetFromOptions(x);
  VecSetFromOptions(b);

  PetscScalar* a;
  VecGetArray(b, &a);
  for (int i=0; i < m; i++)
    a[i] = 0.01;
  VecRestoreArray(b, &a);

  KSP ksp;
  KSPCreate(PETSC_COMM_WORLD, &ksp);
  KSPSetOperators(ksp, A, A);
  KSPSetFromOptions(ksp);

  printf("KSP solve...");
  double t_start = getTime();
  KSPSolve(ksp, b, x);
  double t_end = getTime();

  printf("finished (%e seconds)\n", t_end - t_start);

  VecDestroy(&x); VecDestroy(&b);
  MatDestroy(&A); KSPDestroy(&ksp);
}



static char help[] = "Solves linear system with UMFPack and Petsc.\n\n";

int main(int argc, char** argv)
{
  PetscErrorCode ierr;

  ierr = PetscInitialize(&argc, &argv, NULL, help);
  if (ierr)
    return ierr;

  printf("First UMFPack solve\n");
  solve_suitesparse();
  printf("\n\nFirst Petsc solve\n");
  solve_petsc();
  printf("\n\nSecond UMFPack solve\n");
  solve_suitesparse();


  ierr = PetscFinalize();
  return ierr;
}


